import polars as pl
from polars._plr import PySeries

__version__: str

def get_crs_from_code(srid: int) -> str | None: ...
def get_crs_authority(definition: str) -> tuple[str, str] | None: ...
def to_python_dict(series: PySeries) -> pl.Series: ...
